"""CEI XLS reader."""
import datetime
import glob
import math
import os
import sys
from typing import List
from typing import Tuple

import pandas as pd
import xlrd

import irpf_cei.b3
import irpf_cei.formatting


IRPF_INVESTIMENT_CODES = {
    "ETF": "74 (ETF)",
    "FII": "73 (FII)",
    "STOCKS": "31 (Ações)",
    "NOT_FOUND": "Não encontrado",
}


def get_xls_filename() -> str:
    """Return first xls filename in current folder or Downloads folder."""
    filenames = glob.glob("InfoCEI*.xls")
    if filenames:
        return filenames[0]
    home = os.path.expanduser("~")
    filenames = glob.glob(os.path.join(home, "Downloads", "InfoCEI*.xls"))
    if filenames:
        return filenames[0]
    return sys.exit(
        "Erro: arquivo não encontrado, confira a documentação para mais informações."
    )


def date_parse(value: str) -> datetime.datetime:
    """Parse dates from CEI report."""
    return datetime.datetime.strptime(value.strip(), "%d/%m/%y")


def validate_period(first: str, second: str) -> int:
    """Consider the year from the first trade date."""
    first_year = int(first[-4:])
    second_year = int(second[-4:])
    if first_year <= second_year and first_year > 2018 and second_year < 2021:
        return second_year
    return sys.exit(
        f"Erro: o período de {first} a {second} não é válido, favor verificar "
        "instruções na documentação."
    )


def validate_header(filepath: str) -> Tuple[int, str]:
    """Validate file header.

    Arguments:
        filepath: CEI report's full path

    Returns:
        Tuple[int, str]: reference year for the report and institution name if valid.
    """
    try:
        basic_df = pd.read_excel(
            filepath,
            usecols="B",
            date_parser=date_parse,
            skiprows=4,
        )
    # exits if empty
    except (ValueError, xlrd.XLRDError):
        sys.exit(
            f"Erro: arquivo {filepath} não se encontra íntegro ou no formato de "
            "relatórios do CEI."
        )

    periods = basic_df["Período de"].iloc[0].split(" a ")
    ref_year = validate_period(periods[0], periods[1])

    instutition = basic_df["Período de"].iloc[4]
    return ref_year, instutition


def read_xls(filename: str) -> pd.DataFrame:
    """Read xls.

    Args:
        filename (str): name of XLS file.

    Returns:
        pd.DataFrame: content of the file.
    """
    df = pd.read_excel(
        filename,
        usecols="B:K",
        parse_dates=["Data Negócio"],
        date_parser=date_parse,
        skipfooter=4,
        skiprows=10,
    )
    return df


# Source: https://realpython.com/python-rounding/
def round_down_money(n: float, decimals: int = 2) -> float:
    """Round float on second decimal cases.

    Args:
        n (float): number.
        decimals (int): Number of decimal cases. Defaults to 2.

    Returns:
        float: rounded number.
    """
    multiplier = 10 ** decimals
    # Type-hint for floor won't work until Python 3.9
    # https://github.com/python/typeshed/issues/3195
    return math.floor(n * multiplier) / multiplier  # type: ignore


def clean_table_cols(source_df: pd.DataFrame) -> pd.DataFrame:
    """Drop columns without values.

    Args:
        source_df (pd.DataFrame): full columns DataFrame.

    Returns:
        pd.DataFrame: DataFrame without columns with no value.
    """
    return source_df.dropna(axis="columns", how="all")


def get_trades(df: pd.DataFrame) -> List[Tuple[str, int]]:
    """Return trades representations.

    Args:
        df (pd.DataFrame): trades DataFrame.

    Returns:
        trades: list of df indexes and string representations.
    """
    df["total_cost_rs"] = df["Valor Total (R$)"].apply(
        lambda x: "R$ " + str(f"{x:.2f}".replace(".", ","))
    )
    df = df.drop(columns=["Valor Total (R$)"])
    list_of_list = df.astype(str).values.tolist()
    df = df.drop(columns=["total_cost_rs"])
    return [(" ".join(x), i) for i, x in enumerate(list_of_list)]


def group_trades(df: pd.DataFrame) -> pd.DataFrame:
    """Group trades by day, asset and action.

    Args:
        df (pd.DataFrame): ungrouped trades.

    Returns:
        pd.DataFrame: grouped trades.
    """
    return (
        df.groupby(["Data Negócio", "Código", "C/V"])
        .agg(
            {
                "Quantidade": "sum",
                "Valor Total (R$)": "sum",
                "Especificação do Ativo": "first",
            }
        )
        .reset_index()
    )


def calculate_taxes(df: pd.DataFrame, auction_trades: List[int]) -> pd.DataFrame:
    """Calculate emolumentos and liquidação taxes based on reference year.

    Args:
        df (pd.DataFrame): grouped trades.
        auction_trades (List[int]): list of auction trades.

    Returns:
        pd.DataFrame: trades with two new columns of calculated taxes.
    """
    df["Liquidação (R$)"] = (
        df["Valor Total (R$)"] * irpf_cei.b3.get_trading_rate()
    ).apply(round_down_money)
    df["Emolumentos (R$)"] = (
        df["Valor Total (R$)"]
        * irpf_cei.b3.get_emoluments_rates(df["Data Negócio"].array, auction_trades)
    ).apply(round_down_money)
    return df


def buy_sell_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Create columns for buys and sells with quantity and total value.

    Args:
        df (pd.DataFrame): grouped trades.

    Returns:
        pd.DataFrame: grouped trades with four new columns of buys and sells.
    """
    df["Quantidade Compra"] = df["Quantidade"].where(df["C/V"].str.contains("C"), 0)
    df["Custo Total Compra (R$)"] = (
        df[["Valor Total (R$)", "Liquidação (R$)", "Emolumentos (R$)"]]
        .sum(axis="columns")
        .where(df["C/V"].str.contains("C"), 0)
    ).round(decimals=2)
    df["Quantidade Venda"] = df["Quantidade"].where(df["C/V"].str.contains("V"), 0)
    df["Custo Total Venda (R$)"] = (
        df[["Valor Total (R$)", "Liquidação (R$)", "Emolumentos (R$)"]]
        .sum(axis="columns")
        .where(df["C/V"].str.contains("V"), 0)
    ).round(decimals=2)
    df.drop(["Quantidade", "Valor Total (R$)"], axis="columns", inplace=True)
    return df


def group_buys_sells(df: pd.DataFrame) -> pd.DataFrame:
    """Group buys and sells by asset.

    Args:
        df (pd.DataFrame): ungrouped buys and sells.

    Returns:
        pd.DataFrame: grouped buys and sells.
    """
    return (
        df.groupby(["Código"])
        .agg(
            {
                "Quantidade Compra": "sum",
                "Custo Total Compra (R$)": "sum",
                "Quantidade Venda": "sum",
                "Custo Total Venda (R$)": "sum",
                "Especificação do Ativo": "first",
            }
        )
        .round(decimals=2)
        .reset_index()
    )


def average_price(df: pd.DataFrame) -> pd.DataFrame:
    """Compute average price.

    Args:
        df (pd.DataFrame): buys and sells without average price.

    Returns:
        pd.DataFrame: buys and sells with average price.
    """
    df["Preço Médio (R$)"] = df["Custo Total Compra (R$)"] / df["Quantidade Compra"]
    return df


def goods_and_rights(source_df: pd.DataFrame) -> pd.DataFrame:
    """Call methods for goods and rights.

    Args:
        source_df (pd.DataFrame): raw DataFrame.

    Returns:
        pd.DataFrame: goods and rights DataFrame.
    """
    result_df = buy_sell_columns(source_df)
    result_df = group_buys_sells(source_df)
    result_df = average_price(result_df)
    return result_df


def output_taxes(tax_df: pd.DataFrame) -> None:
    """Print tax DataFrame.

    Args:
        tax_df (pd.DataFrame): calculated tax columns.
    """
    with pd.option_context("display.max_rows", None, "display.max_columns", None):
        print("Valores calculados de emolumentos, liquidação e custo total:\n", tax_df)


def output_goods_and_rights(
    result_df: pd.DataFrame, ref_year: int, institution: str
) -> None:
    """Return a list of assets."""
    pd.set_option("float_format", irpf_cei.formatting.get_currency_format())
    print("========= Bens e Direitos =========")
    for row in result_df.iterrows():
        idx = row[0]
        content = row[1]
        desc = content["Especificação do Ativo"]
        code = content["Código"]
        qtd = content["Quantidade Compra"] - content["Quantidade Venda"]
        avg_price = content["Preço Médio (R$)"]
        avg_price_str = irpf_cei.formatting.fmt_money(avg_price, 3)
        cnpj = irpf_cei.b3.get_cnpj_institution(institution)
        result = irpf_cei.formatting.fmt_money(avg_price * qtd, 2)
        asset_info = irpf_cei.b3.get_asset_info(code)
        print(
            f"============= Ativo {idx + 1} =============\n"
            f"Código: {IRPF_INVESTIMENT_CODES[asset_info.category]}\n"
            f"CNPJ: {asset_info.cnpj if asset_info.cnpj else 'Não encontrado'}\n"
            f"Discriminação (sugerida): {desc}, código: {code}, quantidade: {qtd}, "
            f"preço médio de compra: R$ {avg_price_str}, corretora: {institution} -"
            f" CNPJ {cnpj}\nSituação em 31/12/{ref_year}: R$ {result}\n"
        )



"""Response objects."""
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union


class ResponseTypes:
    """Response types class."""

    PARAMETERS_ERROR = "ParametersError"
    RESOURCE_ERROR = "ResourceError"
    SYSTEM_ERROR = "SystemError"
    SUCCESS = "Success"


class ResponseFailure:
    """Response failure class."""

    def __init__(self, type_: str, message: Union[str, Exception, None]) -> None:
        """Constructor."""
        self.type = type_
        self.message = self._format_message(message)

    def _format_message(self, msg: Union[str, Exception, None]) -> Optional[str]:
        """Format message when it is an exception.

        Args:
            msg: string, exception or None.

        Returns:
            Union[str, None]: formatted message or None.
        """
        if isinstance(msg, Exception):
            return f"{msg.__class__.__name__}: {msg}"
        return msg

    @property
    def value(self) -> Dict[str, Optional[str]]:
        """Value property.

        Returns:
            Dict[str, str]: type and message.
        """
        return {"type": self.type, "message": self.message}

    def __bool__(self) -> bool:
        """Bool return for success."""
        return False


class ResponseSuccess:
    """Response success class."""

    def __init__(self, value: Any = None) -> None:
        """Constructor."""
        self.type = ResponseTypes.SUCCESS
        self.value = value

    def __bool__(self) -> bool:
        """Bool return for success."""
        return True



"""Command-line interface."""
from typing import Any
from typing import List
from typing import Tuple

import click
import inquirer

import irpf_cei.cei
import irpf_cei.formatting


def select_trades(trades: List[Tuple[str, int]]) -> Any:
    """Checkbox selection of auction trades.

    Args:
        trades (List[Tuple[str, int]]): list of all trades and indexes.

    Returns:
        Any: list of indexes of selected auction trades.
    """
    click.secho(
        (
            "Para o cálculo dos emolumentos é necessário informar operações"
            "realizadas em horário de leilão. Essa informação é obtida com "
            "a sua corretora através de relatórios de ordem de compra."
        ),
        fg="green",
    )
    while True:
        selection = inquirer.prompt(
            [
                inquirer.Checkbox(
                    "trades",
                    message=(
                        "Quais operações foram realizadas em horário de leilão? "
                        "(Selecione apertando espaço e ao terminar aperte enter)"
                    ),
                    choices=trades,
                )
            ]
        )["trades"]
        if len(selection) == 0:
            answer = inquirer.prompt(
                [
                    inquirer.List(
                        "",
                        message="Nenhuma operação selecionada.\nIsso está correto?",
                        choices=["Sim", "Não"],
                    )
                ]
            )[""]
            if answer == "Sim":
                return []
        else:
            return selection


@click.command()
@click.version_option()
def main() -> None:
    """Sequecence of operations for trades."""
    response = irpf_cei.formatting.set_pt_br_locale()
    if not response:
        click.secho(
            f"Erro: {response.value['message']}",
            fg="red",
            err=True,
        )
        # Raises SystemExit
        raise click.ClickException("")
    filename = irpf_cei.cei.get_xls_filename()
    click.secho(f"Nome do arquivo: {filename}", fg="blue")

    ref_year, institution = irpf_cei.cei.validate_header(filename)
    source_df = irpf_cei.cei.read_xls(filename)
    source_df = irpf_cei.cei.clean_table_cols(source_df)
    source_df = irpf_cei.cei.group_trades(source_df)
    trades = irpf_cei.cei.get_trades(source_df)
    auction_trades = select_trades(trades)
    tax_df = irpf_cei.cei.calculate_taxes(source_df, auction_trades)
    irpf_cei.cei.output_taxes(tax_df)
    result_df = irpf_cei.cei.goods_and_rights(tax_df)
    irpf_cei.cei.output_goods_and_rights(result_df, ref_year, institution)


if __name__ == "__main__":
    main()  # pragma: no cover



"""Formatting module."""
import locale
import math
from typing import Any
from typing import Callable
from typing import Union

import irpf_cei.responses as res


def set_pt_br_locale() -> Union[res.ResponseFailure, res.ResponseSuccess]:
    """Sets pt_BR locale."""
    # one gets available locale from shell `locale -a`
    supported_locales = ["pt_BR.utf8", "pt_BR.UTF-8"]
    for loc in supported_locales:
        try:
            locale.setlocale(locale.LC_ALL, loc)
            return res.ResponseSuccess()
        except locale.Error:
            pass
    return res.ResponseFailure(
        res.ResponseTypes.SYSTEM_ERROR,
        "locale pt_BR não encontrado, confira a documentação para mais informações.",
    )


def get_currency_format() -> Callable[[Any], str]:
    """Return currency function.

    Returns:
        Callable[[Any], str]: function from current locale.
    """
    return locale.currency


def fmt_money(amount: float, ndigits: int = 2) -> str:
    """Return padded and rounded value."""
    if math.isnan(amount):
        return "N/A"
    rounded = round(amount, ndigits)
    result = str(rounded).replace(".", ",")
    rounded_digits = result.split(",")[1]
    missing_digits = ndigits - len(rounded_digits)
    padded_result = result + "0" * missing_digits
    return padded_result



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['irpf_cei']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.1,<9.0.0',
 'inquirer>=2.8.0,<3.0.0',
 'pandas>=1.3.0,<2.0.0',
 'xlrd>=1.2,<3.0']

entry_points = \
{'console_scripts': ['irpf-cei = irpf_cei.__main__:main']}

setup_kwargs = {
    'name': 'irpf-cei',
    'version': '3.5.2',
    'description': 'Programa auxiliar gratuito para calcular custos de ações, ETFs e fundos imobiliários.',
    'long_description': 'IRPF CEI\n========\n\n**NOTA: este programa foi completamente substituído pelo** `IRPF Investidor`_ **e não será mais atualizado.**\n\n.. _IRPF Investidor: https://github.com/staticdev/irpf-investidor/\n',
    'author': 'staticdev',
    'author_email': 'staticdev-support@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/staticdev/irpf-cei',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
